"""Sequence class.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
"""
import typing as t
from .elements import Element, NamedElement


class Sequence(NamedElement):

    __slots__ = ('_elements',)

    def __init__(self, *elements: t.Union[Element, str]):
        self._elements = self._validate_elements(elements)

    def _get_node_result(self, root, tree, rule, s, node):
        pos = node.start
        is_valid = True

        for elem in self._elements:
            is_valid, pos = root._walk(elem, pos, node.children, rule, True)
            if not is_valid:
                return is_valid, pos

        root._append_tree(tree, node, pos)

        return is_valid, pos

    def _run_export_js(self, js_indent, indent, classes, cname):
        return self._export_js_elements(js_indent, indent, classes, cname)

    def _run_export_py(self, py_indent, indent, classes):
        return self._export_py_elements(py_indent, indent, classes)

    def _run_export_c(self, c_indent, indent, enums):  # type: ignore
        return self._export_c_elements(c_indent, indent, enums)  # type: ignore

    def _run_export_go(self, go_indent, indent, enums):
        return self._export_go_elements(go_indent, indent, enums)

    def _run_export_java(self, java_indent, indent, enums, classes):
        return self._export_java_elements(java_indent, indent, enums, classes)



"""Element and NamedElement Class.

These are the base classes used for all other elements.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
"""
import typing as t
from abc import abstractmethod
if t.TYPE_CHECKING:
    from .grammar import Grammar
    from .node import Node


def camel_case(s: str) -> str:
    return ''.join(
        p[0].upper() + p[1:] if n else p
        for n, p in enumerate(s.split('_')))


def cap_case(s: str) -> str:
    return ''.join(p[0].upper() + p[1:] for p in s.split('_') if p)


_cref: t.Optional['Element'] = None
T = t.TypeVar('T', bound='Element')


def c_export(func: t.Callable[
            [T, str, int, t.Set[str], str],
            str
        ]) -> t.Callable[
            [T, str, int, t.Set[str], t.Optional['Element']],
            str
        ]:

    def wrapper(self, c_indent, indent, enums, ref=None):
        global _cref
        elem, _cref = self if _cref is None else _cref, ref
        gid = getattr(elem, 'name', getattr(elem, '_name', 'CLERI_NONE'))
        if gid != 'CLERI_NONE':
            gid = 'CLERI_GID_{}'.format(gid.upper())
            enums.add(gid)

        return func(self, c_indent, indent, enums, gid)

    return wrapper


def go_export(func: t.Callable[
            [T, str, int, t.Set[str], str],
            str
        ]) -> t.Callable[
            [T, str, int, t.Set[str]],
            str
        ]:
    def wrapper(self, go_indent, indent, enums):
        gid = getattr(self, 'name', getattr(self, '_name', 'NoGid'))
        if gid != 'NoGid':
            gid = 'Gid{}'.format(cap_case(gid))
            enums.add(gid)
        return func(self, go_indent, indent, enums, gid)

    return wrapper


def java_export(func: t.Callable[
            [T, str, int, t.Set[str], t.Set[str], t.Optional[str]],
            str
        ]) -> t.Callable[
            [T, str, int, t.Set[str], t.Set[str]],
            str
        ]:

    def wrapper(self, java_indent, indent, enums, classes):
        classes.add('jleri.{}'.format(self.__class__.__name__.lstrip('_')))
        gid = getattr(self, 'name', getattr(self, '_name', None))
        if gid is not None:
            assert isinstance(gid, str)
            gid = gid.upper()
            enums.add(gid)
        return func(self, java_indent, indent, enums, classes, gid)

    return wrapper


class Element:

    __slots__ = tuple()
    name: str
    _name: str

    @staticmethod
    def _validate_element(element: t.Union[str, 'Element']) -> 'Element':
        if isinstance(element, str):
            return Token(element)
        if isinstance(element, Element):
            return element
        raise TypeError(
            'Expecting an element or string '
            'but received type: {}'.format(type(element)))

    @classmethod
    def _validate_elements(cls,
                           elements: t.Tuple[t.Union[str, 'Element'], ...]
                           ) -> t.List['Element']:
        return [cls._validate_element(elem) for elem in elements]

    @abstractmethod
    def _get_node_result(self,
                         root: 'Grammar',
                         tree: t.List['Node'],
                         rule: 'Element',
                         s: str,
                         node: 'Node') -> t.Tuple[bool, int]:
        pass

    def _export_py(self, *args, **kwargs) -> str:
        return "must_be_called_on_named_element"

    def _export_js(self, *args, **kwargs) -> str:
        return "must_be_called_on_named_element"

    def _export_c(self, *args, **kwargs) -> str:
        return "must_be_called_on_named_element"

    def _export_go(self, *args, **kwargs) -> str:
        return "must_be_called_on_named_element"

    def _export_java(self, *args, **kwargs) -> str:
        return "must_be_called_on_named_element"


class NamedElement(Element):

    __slots__ = ('name',)

    def _export_js(
            self,
            js_indent: str,
            indent: int,
            classes: t.Set[str],
            cname: t.Optional[str]) -> str:
        classes.add(self.__class__.__name__.lstrip('_'))
        if hasattr(self, 'name') and indent > 0:
            return '{}.{}'.format(cname, self.name) if cname else self.name
        indent = 0 if indent < 0 else 1 if indent == 0 else indent
        return self._run_export_js(js_indent, indent, classes, cname)

    def _export_js_elements(
            self,
            js_indent: str,
            indent: int,
            classes: t.Set[str],
            cname: t.Optional[str]) -> str:
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=js_indent * new_indent,
            elem=elem._export_js(
                js_indent,
                new_indent, classes, cname))
            for elem in self._elements])  # type: ignore
        return '{class_name}(\n{value}\n{indent})'.format(
            class_name=self.__class__.__name__.lstrip('_'),
            value=value,
            indent=js_indent * indent)

    def _run_export_js(self,
                       js_indent: str,
                       indent: int,
                       classes: t.Set[str],
                       cname: t.Optional[str]) -> str:
        return 'not_implemented'

    def _export_py(self,
                   py_indent: str,
                   indent: int,
                   classes: t.Set[str]) -> str:
        classes.add(self.__class__.__name__.lstrip('_'))
        if hasattr(self, 'name') and indent:
            return self.name
        return self._run_export_py(py_indent, indent or 1, classes)

    def _export_py_elements(self,
                            py_indent: str,
                            indent: int,
                            classes: t.Set[str]) -> str:
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=py_indent * new_indent,
            elem=elem._export_py(
                py_indent,
                new_indent,
                classes))
            for elem in self._elements])  # type: ignore
        return '{class_name}(\n{value}\n{indent})'.format(
            class_name=self.__class__.__name__.lstrip('_'),
            value=value,
            indent=py_indent * indent)

    def _run_export_py(self,
                       py_indent: str,
                       indent: int,
                       classes: t.Set[str]) -> str:
        return 'not_implemented'

    @c_export
    def _export_c(self,
                  c_indent: str,
                  indent: int,
                  enums: t.Set[str],
                  gid: str) -> str:
        if hasattr(self, 'name') and indent:
            return self.name
        return self._run_export_c(c_indent, indent or 1, enums)  # type: ignore

    @c_export
    def _export_c_elements(self,
                           c_indent: str,
                           indent: int,
                           enums: t.Set[str],
                           gid: str) -> str:
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=c_indent * new_indent,
            elem=elem._export_c(
                c_indent,
                new_indent,
                enums))
            for elem in self._elements])  # type: ignore
        return \
            'cleri_{class_name}(\n{gid},\n{num},\n{value}\n{indent})'.format(
                class_name=self.__class__.__name__.lstrip('_').lower(),
                gid='{indent}{gid}'.format(
                    indent=c_indent * (indent + 1),
                    gid=gid),
                num='{indent}{num}'.format(
                    indent=c_indent * (indent + 1),
                    num=len(self._elements)),  # type: ignore
                value=value,
                indent=c_indent * indent)

    @c_export
    def _run_export_c(self,
                      c_indent: str,
                      indent: int,
                      enums: t.Set[str],
                      gid: str) -> str:
        return 'not_implemented'

    @go_export
    def _export_go(self,
                   go_indent: str,
                   indent: int,
                   enums: t.Set[str],
                   gid: str) -> str:
        if hasattr(self, 'name') and indent:
            return camel_case(self.name)
        return self._run_export_go(go_indent, indent or 1, enums)

    @go_export
    def _export_go_elements(self,
                            go_indent: str,
                            indent: int,
                            enums: t.Set[str],
                            gid: str) -> str:
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=go_indent * new_indent,
            elem=elem._export_go(
                go_indent,
                new_indent,
                enums)) for elem in self._elements])  # type: ignore
        return '{class_name}(\n{gid},\n{value},\n{indent})'.format(
            class_name='goleri.New' + self.__class__.__name__.lstrip('_'),
            gid='{indent}{gid}'.format(
                indent=go_indent * (indent + 1),
                gid=gid),
            value=value,
            indent=go_indent * indent)

    @go_export
    def _run_export_go(self,
                       go_indent: str,
                       indent: int,
                       enums: t.Set[str],
                       gid: str) -> str:
        return 'not_implemented'

    @java_export
    def _export_java(self,
                     java_indent: str,
                     indent: int,
                     enums: t.Set[str],
                     classes: t.Set[str],
                     gid: t.Optional[str]) -> str:
        if hasattr(self, 'name') and indent > 0:
            return self.name.upper()
        return self._run_export_java(
            java_indent, abs(indent) or 1, enums, classes)

    @java_export
    def _export_java_elements(self,
                              java_indent: str,
                              indent: int,
                              enums: t.Set[str],
                              classes: t.Set[str],
                              gid: t.Optional[str]) -> str:
        new_indent = indent + 1
        value = ',\n'.join(['{indent}{elem}'.format(
            indent=java_indent * new_indent,
            elem=elem._export_java(
                java_indent,
                new_indent,
                enums,
                classes)) for elem in self._elements])  # type: ignore
        return '{class_name}({gid}\n{value}\n{indent})'.format(
            class_name='new ' + self.__class__.__name__.lstrip('_'),
            gid='' if gid is None else '\n{indent}Ids.{gid},'.format(
                indent=java_indent * (indent + 1),
                gid=gid),
            value=value,
            indent=java_indent * indent)

    @java_export
    def _run_export_java(self,
                         java_indent: str,
                         indent: int,
                         enums: t.Set[str],
                         classes: t.Set[str],
                         gid: t.Optional[str]):
        return 'not_implemented'

# Added this import to the bottom to prevent circular import cycle.
# Note: usually this is bad design but in this case we do want class
#       inheritance which allows us to create a new class Token which
#       is sub-classed from the 'NamedElement' class.


from .token import Token  # nopep8  # noqa: E402



"""JSON Grammar."""
from pyleri import (
    Ref,
    Choice,
    Grammar,
    Regex,
    Keyword,
    Sequence,
    List)


class JsonGrammar(Grammar):
    START = Ref()

    # JSON strings should be enclosed in double quotes.
    # A backslash can be used as escape character.
    r_string = Regex(r'(")(?:(?=(\\?))\2.)*?\1')

    # JSON does not support floats or integers prefixed with a + sign
    # and floats must start with a number, for example .5 is not allowed
    # but should be written like 0.5
    r_float = Regex(r'-?[0-9]+\.?[0-9]+')
    r_integer = Regex('-?[0-9]+')

    k_true = Keyword('true')
    k_false = Keyword('false')
    k_null = Keyword('null')

    json_map_item = Sequence(r_string, ':', START)

    json_map = Sequence('{', List(json_map_item), '}')
    json_array = Sequence('[', List(START), ']')

    START = Choice(
        r_string,
        r_float,
        r_integer,
        k_true,
        k_false,
        k_null,
        json_map,
        json_array)



"""pyleri.Keyword Class.

Try matching a given keyword string. The keyword should match
Grammer.RE_KEYWORDS otherwise the keyword will not be found. It's possible
however to overwrite the default RE_KEYWORDS in you own Grammar class.

:copyright: 2021, Jeroen van der Heijden <jeroen@cesbit.com>
"""
import typing as t
from .elements import NamedElement, c_export, go_export, java_export


class Keyword(NamedElement):

    __slots__ = ('_keyword', '_ign_case')

    def __init__(self, keyword: str, ign_case: bool = False):

        if not isinstance(keyword, str):
            raise TypeError(
                'Keyword(): first positional argument must be a string value')

        self._keyword = keyword
        self._ign_case = bool(ign_case)

    @property
    def ign_case(self) -> bool:
        return self._ign_case

    def __repr__(self):
        return self._keyword

    def _get_node_result(self, root, tree, rule, s, node):
        try:
            re_match = root._cached_kw_match[node.start]
        except KeyError:
            re_match = \
                root._cached_kw_match[node.start] = \
                root.RE_KEYWORDS.match(s)
        is_valid = bool(
            re_match and
            (re_match.group(0) == self._keyword or
             (self._ign_case and
              re_match.group(0).lower() == self._keyword.lower())))
        if is_valid:
            root._append_tree(tree, node, node.start + len(self._keyword))
        else:
            root._expecting.update(self, node.start)

        return is_valid, node.end or node.start

    def _run_export_js(self, js_indent, indent, classes, cname):
        return 'Keyword(\'{}\'{})'.format(
            self._keyword.replace('\'', '\\\''),
            ', true' if self._ign_case else '')

    def _run_export_py(self, py_indent, indent, classes):
        return 'Keyword(\'{}\'{})'.format(
            self._keyword.replace('\'', '\\\''),
            ', True' if self._ign_case else '')

    @c_export
    def _run_export_c(self, c_indent, indent, enums, gid):
        return 'cleri_keyword({}, "{}", {})'.format(
            gid,
            self._keyword.replace('"', '\\"'),
            'CLERI_CASE_INSENSITIVE'
            if self._ign_case else 'CLERI_CASE_SENSITIVE')

    @go_export
    def _run_export_go(self, go_indent, indent, enums, gid):
        return 'goleri.NewKeyword({}, "{}", {})'.format(
            gid,
            self._keyword.replace('"', '\\"'),
            'true' if self._ign_case else 'false')

    @java_export
    def _run_export_java(self, c_indent, indent, enums, classes, gid):
        return 'new Keyword({}"{}", {})'.format(
            '' if gid is None else 'Ids.{}, '.format(gid),
            self._keyword.replace('"', '\\"'),
            'true' if self._ign_case else 'false')
